﻿// WIN32API による中央精機製ステージコントローラ QT-A / QT-B と通信するためのサンプルです。
// サンプルのため最低限の記述を心がけています。
// エラー処理、実装方法など開発言語についてはお使いの開発環境のヘルプなどをご参照いただくか、開発言語メーカへお問い合わせください。
// 
// 【動作確認環境】
// ・ OS: Windows 10 22H2 / Windows 11 23H2
// ・ Visual C++ 2022
// 
// 【前提条件】
// ・ このサンプルではコントローラの A 軸を動作させます。
// ・ コントローラとステージが正しく接続され、正常に動作する設定になっていることとします。
// ・ コントローラはレスポンスなし (X:0) を想定しています。
// 
// 【サンプルコードの内容】
// 1. 通信ポートを開く
// 2. 通信ポートの設定を行う
// 3. QT-A / QT-B との RS-232C 通信を行う
//    - A 軸に接続されたステージを +1000 パルス移動開始します。
//    - 100 ms 待ちます。
//    - Q コマンドを送信し、偏とされた内容をコンソールに出力します。
// 4. 通信ポートを閉じる
// 
// 2022/10/28 時点では Microsoft アプリ開発では「通信リソース」についてのドキュメントが公開されています。
// https://learn.microsoft.com/ja-jp/windows/win32/devio/communications-resources
// 
// 【使用方法】
// 1. 通信ポート設定を定数 (constexpr) PORT_NUMBER、BAUDRATE、DATABITS、PARITY、STOPBITS に行います。
// 2. コントローラに設定しているデリミタに合わせ DELIMITER を変更します。

#include <stdio.h>
#include <windows.h>

// <<<<< 設定箇所 - ここから >>>>>
// 通信ポートの番号を指定します。
// 通信ポート番号はデバイスマネージャなどで確認します。
const int PORT_NUMBER = 3;

// 通信ポートのボーレート 4800, 9600, 19200, (QT-B のみ) 38400 の何れかを指定します。
// お使いのコントローラの設定 (パラメータ No.53) に合わせてください。
const DWORD BAUDRATE = CBR_9600;
//const auto BAUDRATE = CBR_4800;
//const auto BAUDRATE = CBR_19200;
//const auto BAUDRATE = CBR_38400; // QT-B のみ

// 通信ポートのデータビット長 8 か 7 bit を指定する。
// お使いのコントローラの設定 (パラメータ No.53) に合わせてください。
const BYTE DATABITS = 8;
//const BYTE DATABITS = 7;

// 通信ポートのパリティ NOPARITY (パリティなし),　PARITY_ODD (奇数パリティ), PARITY_EVEN (偶数パリティ) の何れかを指定する。
// お使いのコントローラの設定 (パラメータ No.53) に合わせてください。
const BYTE PARITY = NOPARITY;
//const BYTE PARITY = ODDPARITY;
//const BYTE PARITY = EVENPARITY;

// 通信ポートのストップビット長 ONESTOPBIT, TWOSTOPBITS の何れかを指定する。
// お使いのコントローラの設定 (パラメータ No.53) に合わせてください。
const BYTE STOPBITS = ONESTOPBIT;
//const BYTE STOPBITS = TWOSTOPBITS;

// コントローラとの通信で使用するデリミタを設定します。
// お使いのコントローラの設定 (パラメータ No.51) に合わせてください。
// 0: CR+LF、1: CR、2: LF
const int DELIMITER = 0;
//const int DELIMITER = 1;
//const int DELIMITER = 2;

// <<<<< 設定箇所 - ここまで >>>>>



#if DELIMITER == 0
const char* DELIMITER_STRING = "\r\n";
#elif DELIMITER == 1
const char* DELIMITER_STRING = "\r";
#elif DELIMITER == 2
const char* DELIMITER_STRING = "\n";
#else
#error "定義 DELMITER はコントローラの設定に合わせ 0: CR+LF、1: CR、2: LF のいずれかを指定してください。"
#endif

// portNumber で指定されたシリアル通信を開きます。
static HANDLE OpenSerialPort(int nPortNumber) {
    wchar_t szPortName[16];
    // CreateFile 関数で使用する通信ポート名は COM1 ～ COM9 ではそのまま COM1 ～ COM9 で指定できますが、
    // COM10 以上では \\.\COM10 と COM の前に追加の文字列が必要となります。
    // ここでは、COM1 ～ 9 も含め \\.\COMx で指定しています。
    swprintf_s(szPortName, 16, L"\\\\.\\COM%d", nPortNumber);
    HANDLE port = CreateFile(szPortName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL); // 通信ポートを開きます。

    return port;
}

// 通信ポートの初期化を行う。
static void InitializeSerialPort(HANDLE hPort) {
    DCB dcb;
    GetCommState(hPort, &dcb);   // 現在の設定値を読み込む

    // お使いの通信ポートによっては追加の設定が必要になります
    // 詳細: https://learn.microsoft.com/ja-jp/windows/win32/api/winbase/ns-winbase-dcb
    dcb.BaudRate = BAUDRATE;
    dcb.ByteSize = DATABITS;
    dcb.Parity = PARITY;
    dcb.fParity = PARITY != NOPARITY;
    dcb.StopBits = STOPBITS;
    dcb.fOutxCtsFlow = FALSE;
    dcb.fOutxDsrFlow = FALSE;
    dcb.fDtrControl = DTR_CONTROL_DISABLE;
    dcb.fRtsControl = RTS_CONTROL_DISABLE;

    SetCommState(hPort, &dcb);

    COMMTIMEOUTS cto;
    GetCommTimeouts(hPort, &cto);

    // 受信タイムアウト [ms] = ReadTotalTimeoutConstant + 文字数× ReadTotalTimeoutMultiplier
    cto.ReadTotalTimeoutMultiplier = 0;
    cto.ReadTotalTimeoutConstant = 1000;
    // 送信タイムアウト [ms] = WriteTotalTimeoutConstant + 文字数× WriteTotalTimeoutMultiplier
    cto.WriteTotalTimeoutMultiplier = 0;
    cto.WriteTotalTimeoutConstant = 1000;

    SetCommTimeouts(hPort, &cto);
}

// 指定した通信ポートからデリミタまで受信します。
// lpBuffer は受信バッファへのポインタです。
// nNumberOfBytesToRead は受信バッファ lpBuffer のバイト数です。
// 
// [ヒント]
// データ受信はこの関数のように 1 文字ずつデリミタを検査する方法のほか、
// 予め受信バイト数が分かる場合は ClearCommError 関数を利用し受信バイト数を監視し、まとめて読み込む方法があります。
static BOOL ReadLine(HANDLE hPort, LPVOID lpBuffer, int nBufferSize) {
    bool eol = false;
    char* buffer = (char*)lpBuffer;
    int i = 0;
    DWORD dwReadSize; // ReadFile 関数を実行すると実際に受信したサイズが書き込まれます。
    while (eol == false) {
        char c = '\0';
        // 通信ポートからの受信は次の ReadFile 関数で行います。
        if (ReadFile(hPort, &c, 1, &dwReadSize, NULL)) {
            buffer[i] = c;
            if (1 < i) {
#if DELIMITER == 0
                if ((buffer[i - 1] == '\r') && (buffer[i] == '\n')) {
                    buffer[i - 1] = '\0';
#elif DELIMITER == 1
                if (buffer[i] == '\r') {
                    buffer[i] = '\0';
#elif DELIMITER == 2
                if (buffer[i] == '\n') {
                    buffer[i] = '\0';
#else
#error "定義 DELMITER はコントローラの設定に合わせ 0: CR+LF、1: CR、2: LF のいずれかを指定してください。"
#endif
                    eol = true;
                } else if (nBufferSize - 1 <= i) {
                    buffer[nBufferSize - 1] = '\0';
                    printf("\r\n通信ポート COM%d から受信した文字数が多すぎます。処理を中断します。\r\n受信した文字は %s です。", PORT_NUMBER, buffer);
                    return false;
                }
            }
        }
        i++;
    }

    printf("\r\n通信ポート COM%d からデータ %s を受信しました。", PORT_NUMBER, buffer);

    return true;
}

int main() {
    // シリアル通信ポートを開きます。
    HANDLE hPort = OpenSerialPort(PORT_NUMBER);
    printf("\r\n通信ポート COM%d のオープンに", PORT_NUMBER);
    if (hPort == INVALID_HANDLE_VALUE) {
        puts("失敗しました。");
        return -1;
    } else {
        puts("成功しました。");
    }

    // シリアル通信ポートの初期化
    InitializeSerialPort(hPort);

    // ここに送受信の処理を記述します。

    // WriteFile 関数の第 2 引数にはコマンドの文字数を CR+LF もしくは CR まで含めて指定します。
    // コマンドの最後の \r\n はコントローラの設定に合わせて変更してください。
    // CR は \r、 CR+LF は \r\n になります。
    DWORD dwWriteSize; // WriteFile 関数を実行するとこの変数に実際に送信したサイズが書き込まれます。
    // MGO コマンドを送信します。
    char cmd[32] = "MGO:A+1000";
    char cmd_msg[32];
    strncpy_s(cmd_msg, 32, cmd, _TRUNCATE);
    strncat_s(cmd, 32, DELIMITER_STRING, _TRUNCATE);
    WriteFile(hPort, cmd, (DWORD)strlen(cmd), &dwWriteSize, NULL);
    printf("通信ポート COM%d にコマンド %s を送信しました。", PORT_NUMBER, cmd_msg);

    // この部分は移動がある程度進むのを待つためのものなのでコントローラとの通信とは関係ありません。
    Sleep(100);

    // Q コマンドを送信し、返答を受信します。
    char cmd2[8] = "Q:A0";
    strncpy_s(cmd_msg, 32, cmd2, _TRUNCATE);
    strncat_s(cmd2, 8, DELIMITER_STRING, _TRUNCATE);
    WriteFile(hPort, cmd2, (DWORD)strlen(cmd2), &dwWriteSize, NULL);
    printf("\r\n通信ポート COM%d にコマンド %s を送信しました。", PORT_NUMBER, cmd_msg);

    // この部分は受信した文字の処理なのでコントローラとの通信には関係ありません。
    char lpBuffer[256] = "";
    ReadLine(hPort, &lpBuffer, 256);
    lpBuffer[255] = '\0';
    int posA = 0;
    char statA[2] = "\0";
    char formatString[8] = "%9d%1s";
    strncat_s(formatString, 8, DELIMITER_STRING, _TRUNCATE);
    sscanf_s(lpBuffer, formatString, &posA, &statA, 2);
    printf("\r\nA 軸: %d + %s", posA, statA);

    // シリアル通信ポートを閉じます。
    CloseHandle(hPort);
    printf("\r\n通信ポート COM%d をクローズしました。", PORT_NUMBER);

    return 0;
}